@@ -1,19 +1,23 @@
 from collections.abc import Mapping
 from datetime import timedelta
+import json
 import logging
 import os
-from typing import Dict, Iterable, Union
+import sys
+from typing import Dict, Iterable, List, Union
+
+import pandas as pd
 
 from beneath import config
 from beneath.admin.client import AdminClient
 from beneath.checkpointer import Checkpointer, PrefixedCheckpointer
 from beneath.config import DEFAULT_READ_BATCH_SIZE
-from beneath.connection import Connection
+from beneath.connection import Connection, GraphQLError
 from beneath.consumer import Consumer
 from beneath.instance import StreamInstance
 from beneath.job import Job
 from beneath.stream import Stream
-from beneath.utils import ProjectQualifier, StreamQualifier, SubscriptionQualifier
+from beneath.utils import infer_avro, ProjectQualifier, StreamQualifier, SubscriptionQualifier
 from beneath.writer import DryWriter, Writer
 
 
@@ -53,10 +57,7 @@ class Client:
         self.connection = Connection(secret=self._get_secret(secret=secret))
         self.admin = AdminClient(connection=self.connection, dry=dry)
         self.dry = dry
-
-        logging.basicConfig()
-        self.logger = logging.getLogger("beneath")
-        self.logger.setLevel(logging.INFO)
+        self.logger = self._make_default_logger()
 
         if dry:
             self._writer = DryWriter(client=self, max_delay_ms=write_delay_ms)
@@ -82,6 +83,20 @@ class Client:
             raise TypeError("secret must be a string")
         return secret.strip()
 
+    @classmethod
+    def _make_default_logger(cls):
+        h1 = logging.StreamHandler(sys.stdout)
+        h1.setLevel(logging.INFO)
+        h1.addFilter(lambda record: record.levelno <= logging.INFO)
+
+        h2 = logging.StreamHandler(sys.stderr)
+        h2.setLevel(logging.WARNING)
+
+        logging.basicConfig(handlers=[h1, h2])
+        logger = logging.getLogger("beneath")
+        logger.setLevel(logging.INFO)
+        return logger
+
     # FINDING AND STAGING STREAMS
 
     async def find_stream(self, stream_path: str) -> Stream:
@@ -109,6 +124,8 @@ class Client:
         log_retention: timedelta = None,
         index_retention: timedelta = None,
         warehouse_retention: timedelta = None,
+        schema_kind: str = "GraphQL",
+        indexes: str = None,
         update_if_exists: bool = None,
     ) -> Stream:
         """
@@ -127,6 +144,8 @@ class Client:
             retention (timedelta):
                 The amount of time to retain records written to the stream.
                 If not set, records will be stored forever.
+            schema_kind (str):
+                The parser to use for ``schema``. Currently must be "GraphQL" (default).
             update_if_exists (bool):
                 If true and the stream already exists, the provided info will be used to update
                 the stream (only supports non-breaking schema changes) before returning it.
@@ -134,7 +153,7 @@ class Client:
         qualifier = StreamQualifier.from_path(stream_path)
         if self.dry:
             data = await self.admin.streams.compile_schema(
-                schema_kind="GraphQL",
+                schema_kind=schema_kind,
                 schema=schema,
             )
             stream = await Stream._make_dry(
@@ -147,8 +166,9 @@ class Client:
                 organization_name=qualifier.organization,
                 project_name=qualifier.project,
                 stream_name=qualifier.stream,
-                schema_kind="GraphQL",
+                schema_kind=schema_kind,
                 schema=schema,
+                indexes=indexes,
                 description=description,
                 meta=meta,
                 use_index=use_index,
@@ -267,6 +287,128 @@ class Client:
         """ Forces the client to flush buffered writes without stopping """
         await self._writer.force_flush()
 
+    async def write_full(
+        self,
+        stream_path: str,
+        records: Union[Iterable[dict], pd.DataFrame],
+        key: Union[str, List[str]] = None,
+        description: str = None,
+        recreate_on_schema_change=False,
+    ):
+        """
+        Infers a schema, creates a stream, and writes a full dataset to Beneath.
+        Each call will create a new primary version for the stream, and delete the old primary
+        version if/when the write completes succesfully.
+
+        Args:
+            stream_path (str):
+                The (desired) path to the stream in the format of "USERNAME/PROJECT/STREAM".
+                The project must already exist.
+            records (list(dict) | pandas.DataFrame):
+                The full dataset to write, either as a list of records or as a Pandas DataFrame.
+                This function uses ``beneath.infer_avro`` to infer a schema for the stream based
+                on the records.
+            key (str | list(str)):
+                The fields to use as the stream's key. If not set, will default to the dataframe
+                index if ``records`` is a Pandas DataFrame, or add a column of incrementing numbers
+                if ``records`` is a list.
+            description (str):
+                A description for the stream.
+            recreate_on_schema_change (bool):
+                If true, and there's an existing stream at ``stream_path`` with a schema that is
+                incompatible with the inferred schema for ``records``, it will delete the existing
+                stream and create a new one instead of throwing an error. Defaults to false.
+        """
+        if self._start_count == 0:
+            raise Exception("Cannot call write_full because the client is stopped")
+        if len(records) == 0:
+            return
+        # hairy defaults for `key`
+        if isinstance(key, str):
+            key = [key]
+        if not key:
+            if isinstance(records, pd.DataFrame):
+                index_named = None
+                for name in records.index.names:
+                    if index_named is None:
+                        index_named = name is not None
+                    elif index_named == (name is None):
+                        raise ValueError(
+                            "Cannot write DataFrame with mixed null and non-null"
+                            f" index names {records.index.names}"
+                        )
+                if index_named:
+                    key = records.index.names
+                    records = records.reset_index()
+                else:
+                    if len(records.index.names) > 1:
+                        key = [f"key{idx}" for idx, name in enumerate(records.index.names)]
+                    else:
+                        key = ["key"]
+                    records.index.names = key
+                    records = records.reset_index()
+            else:
+                for idx, record in enumerate(records):
+                    if "key" not in record:
+                        record["key"] = idx
+                    key = ["key"]
+
+        # infer schema and indexes
+        schema = infer_avro(records)
+        indexes = json.dumps([{"key": True, "fields": key}])
+
+        # create the stream
+        try:
+            stream = await self.create_stream(
+                stream_path=stream_path,
+                schema=schema,
+                description=description,
+                schema_kind="Avro",
+                indexes=indexes,
+                update_if_exists=True,
+            )
+        except GraphQLError as e:
+            if "Schema error:" not in str(e):
+                raise
+            if not recreate_on_schema_change:
+                raise ValueError(
+                    "Cannot create stream because an existing stream *with a different schema* "
+                    f"already exists at '{stream_path}'. To delete the existing stream and all its"
+                    " versions and records, pass recreate_on_schema_change=True."
+                )
+            # recreate
+            stream = await self.find_stream(stream_path)
+            await self.admin.streams.delete(stream_id=str(stream.stream_id))
+            stream = await self.create_stream(
+                stream_path=stream_path,
+                schema=schema,
+                description=description,
+                schema_kind="Avro",
+                indexes=indexes,
+                update_if_exists=True,
+            )
+
+        # get instance
+        next_instance = None
+        previous_instance = None
+        if stream.primary_instance is None:
+            next_instance = await stream.create_instance()
+        elif stream.primary_instance.is_final:
+            next_instance = await stream.create_instance()
+            previous_instance = stream.primary_instance
+        else:
+            next_instance = stream.primary_instance
+
+        # write records
+        await next_instance.write(records)
+
+        # make primary and final, and delete previous instance
+        await next_instance.update(make_primary=True, make_final=True)
+        if previous_instance:
+            await previous_instance.delete()
+
+        return stream
+
     # CHECKPOINTERS
 
     async def checkpointer(
