@@ -7,9 +7,11 @@ from typing import TYPE_CHECKING
 if TYPE_CHECKING:
     from beneath.client import Client
 
-from typing import Iterable
+from typing import Iterable, Mapping, Union
 import uuid
 
+from beneath import config
+from beneath.cursor import Cursor
 from beneath.instance import StreamInstance
 from beneath.schema import Schema
 from beneath.utils import StreamQualifier
@@ -40,6 +42,15 @@ class Stream:
     This is probably the object you will use to write/query the stream.
     """
 
+    use_log: bool
+    """ Whether log queries are supported for this stream. """
+
+    use_index: bool
+    """ Whether index queries are supported for this stream. """
+
+    use_warehouse: bool
+    """ Whether warehouse queries are supported for this stream. """
+
     _client: Client
     _qualifier: StreamQualifier
 
@@ -69,6 +80,9 @@ class Stream:
                     stream=stream,
                     admin_data=admin_data["primaryStreamInstance"],
                 )
+        stream.use_log = admin_data["useLog"]
+        stream.use_index = admin_data["useIndex"]
+        stream.use_warehouse = admin_data["useWarehouse"]
         return stream
 
     @classmethod
@@ -84,6 +98,9 @@ class Stream:
         stream.stream_id = None
         stream.schema = Schema(avro_schema)
         stream.primary_instance = await stream.create_instance(version=0, make_primary=True)
+        stream.use_log = True
+        stream.use_index = True
+        stream.use_warehouse = True
         return stream
 
     async def _load_admin_data(self):
@@ -93,6 +110,11 @@ class Stream:
             stream_name=self._qualifier.stream,
         )
 
+    # STATE
+
+    def __repr__(self):
+        return f'<beneath.stream.Stream("{config.BENEATH_FRONTEND_HOST}/{self._qualifier}")>'
+
     # INSTANCES
 
     async def find_instances(self) -> Iterable[StreamInstance]:
@@ -122,6 +144,8 @@ class Stream:
             if self.primary_instance and self.primary_instance.version == version:
                 return self.primary_instance
             raise Exception("can't find instance by version for stream created with a dry client")
+        if self.primary_instance and self.primary_instance.version == version:
+            return self.primary_instance
         admin_data = await self._client.admin.streams.find_instance(
             stream_id=str(self.stream_id),
             version=version,
@@ -131,7 +155,7 @@ class Stream:
 
     async def create_instance(
         self,
-        version: int,
+        version: int = None,
         make_primary=None,
         update_if_exists=None,
     ) -> StreamInstance:
@@ -141,7 +165,8 @@ class Stream:
 
         Args:
             version (int):
-                The version number to assign to the instance
+                The version number to assign to the instance. If not set, will create a new instance
+                with a higher version number than any previous instance for the stream.
             make_primary (bool):
                 Immediately make the new instance the stream's primary instance
             update_if_exists (bool):
@@ -161,7 +186,7 @@ class Stream:
             instance = StreamInstance._make_dry(
                 client=self._client,
                 stream=self,
-                version=version,
+                version=(0 if version is None else version),
                 make_primary=make_primary,
             )
         if make_primary:
@@ -178,3 +203,29 @@ class Stream:
         if not self.stream_id:
             raise Exception("cannot delete dry stream")
         await self._client.admin.streams.delete(self.stream_id)
+
+    # CURSORS
+
+    def restore_cursor(self, replay_cursor: bytes, changes_cursor: bytes):
+        """
+        Restores a cursor previously obtained by querying one of the stream's instances.
+        You must provide the cursor bytes, which can be found as properties of the Cursor object.
+        """
+        return Cursor(
+            connection=self._client.connection,
+            schema=self.schema,
+            replay_cursor=replay_cursor,
+            changes_cursor=changes_cursor,
+        )
+
+    # WRITING RECORDS
+
+    async def write(self, records: Union[Mapping, Iterable[Mapping]]):
+        """
+        Writes records to the stream's primary instance.
+
+        This is a convenience wrapper for ``stream.primary_instance.write(...)``.
+        """
+        if not self.primary_instance:
+            raise Exception("cannot write because the stream doesn't have a primary instance")
+        await self.primary_instance.write(records)
